/**
 * Copyright (c) 2020 Obeo
 * This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *  Obeo - initial API and implementation
 */
package fr.obeo.dsl.tuto.mindstorms.layout.options;

import fr.obeo.dsl.tuto.mindstorms.layout.MindstormsLayoutProvider;
import fr.obeo.dsl.tuto.mindstorms.layout.MindstormsMetadataProvider;
import org.eclipse.elk.core.AbstractLayoutProvider;
import org.eclipse.elk.core.data.ILayoutMetaDataProvider;
import org.eclipse.elk.core.data.LayoutAlgorithmData;
import org.eclipse.elk.core.options.CoreOptions;
import org.eclipse.elk.graph.properties.IProperty;
import org.eclipse.elk.graph.properties.Property;

@SuppressWarnings("all")
public class MindstormsOptions implements ILayoutMetaDataProvider {
  /**
   * The id of the Mindstorms algorithm.
   */
  public static final String ALGORITHM_ID = "fr.obeo.dsl.tuto.mindstorms.layout.Mindstorms";
  
  /**
   * Default value for {@link #SPACING_NODE_NODE} with algorithm "Mindstorms".
   */
  private static final double SPACING_NODE_NODE_DEFAULT = 50;
  
  /**
   * The minimal distance to be preserved between each two nodes.
   * <h3>Algorithm Specific Details</h3>
   * Used as horizontal node spacing
   */
  public static final IProperty<Double> SPACING_NODE_NODE = new Property<Double>(
                                CoreOptions.SPACING_NODE_NODE,
                                SPACING_NODE_NODE_DEFAULT);
  
  /**
   * The vertical spacing to be preserved between each line, ie between the lowest node of the upper line and the highest node of the lower line.
   * Note that 'spacing.nodeNode' is used for the spacing between nodes within the line itself.
   */
  public static final IProperty<Double> SPACING_LINE = MindstormsMetadataProvider.SPACING_LINE;
  
  /**
   * The number of nodes to display per line.
   */
  public static final IProperty<Double> NODE_PLACEMENT_NB_NODES_PER_LINE = MindstormsMetadataProvider.NODE_PLACEMENT_NB_NODES_PER_LINE;
  
  /**
   * Layouter-specific algorithm factory.
   */
  public static class MindstormsFactory implements org.eclipse.elk.core.util.IFactory<AbstractLayoutProvider> {
    public AbstractLayoutProvider create() {
      AbstractLayoutProvider provider = new MindstormsLayoutProvider();
      provider.initialize("");
      return provider;
    }
    
    public void destroy(final AbstractLayoutProvider obj) {
      obj.dispose();
    }
  }
  
  public void apply(final org.eclipse.elk.core.data.ILayoutMetaDataProvider.Registry registry) {
    registry.register(new LayoutAlgorithmData.Builder()
        .id("fr.obeo.dsl.tuto.mindstorms.layout.Mindstorms")
        .name("Mindstorms")
        .description("This layout is dedicated to Choregraphy diagram of Mindstorms Sirius tutorial")
        .providerFactory(new MindstormsFactory())
        .melkBundleName(null)
        .definingBundleId("fr.obeo.dsl.tuto.mindstorms.layout")
        .create()
    );
    registry.addOptionSupport(
        "fr.obeo.dsl.tuto.mindstorms.layout.Mindstorms",
        "org.eclipse.elk.spacing.nodeNode",
        SPACING_NODE_NODE_DEFAULT
    );
    registry.addOptionSupport(
        "fr.obeo.dsl.tuto.mindstorms.layout.Mindstorms",
        "fr.obeo.dsl.tuto.mindstorms.layout.spacing.line",
        SPACING_LINE.getDefault()
    );
    registry.addOptionSupport(
        "fr.obeo.dsl.tuto.mindstorms.layout.Mindstorms",
        "fr.obeo.dsl.tuto.mindstorms.layout.nodePlacement.nbNodesPerLine",
        NODE_PLACEMENT_NB_NODES_PER_LINE.getDefault()
    );
  }
}
