/**
 * Copyright (c) 2020 Obeo
 * This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 * Contributors:
 *  Obeo - initial API and implementation
 */
package fr.obeo.dsl.tuto.mindstorms.layout;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.eclipse.elk.core.AbstractLayoutProvider;
import org.eclipse.elk.core.options.CoreOptions;
import org.eclipse.elk.core.options.EdgeRouting;
import org.eclipse.elk.core.util.IElkProgressMonitor;
import org.eclipse.elk.graph.ElkEdge;
import org.eclipse.elk.graph.ElkEdgeSection;
import org.eclipse.elk.graph.ElkNode;
import org.eclipse.elk.graph.util.ElkGraphUtil;

import fr.obeo.dsl.tuto.mindstorms.layout.options.MindstormsOptions;

/**
 * An example of layout algorithm class adapted to the <a href="https://wiki.eclipse.org/Sirius/Tutorials/Mindstorms">Mindstorms modeler</a>. <BR/>
 * This algorithm:
 * <UL>
 * <LI>aligns nodes on lines.</LI>
 * <LI>The number of nodes per line depends on the option {@link MindstormsOptions#NODE_PLACEMENT_NB_NODES_PER_LINE}.</LI>
 * <LI>The space between 2 lines can be configured with option {@link MindstormsOptions#SPACING_LINE}.</LI>
 * <LI>The space between each node can be configured with option {@link MindstormsOptions#SPACING_NODE_NODE}.</LI>
 * </UL>
 */
public class MindstormsLayoutProvider extends AbstractLayoutProvider {
	/** 
	 * The space between the top of the node and the edge.
	 * Another option could be added to make this value configurable. 
	 */
	public static final double SPACE_FROM_TOP_OF_NODE = 16;

	/** 
	 * The horizontal space to do the loop for the edge at the end of the line and at the start of a new line.
	 * Another option could be added to make this value configurable. 
	 */
	public static final double HORIZONTAL_EGDE_LOOP_SPACING = 20;
	
    @Override
    public void layout(ElkNode layoutGraph, IElkProgressMonitor progressMonitor) {
    	// The layoutGraph contains all the initial location and size of all nodes and edges.  
    	
        // Start progress monitor
        progressMonitor.begin("Mindstorms", 3);
        progressMonitor.log("Algorithm began");

        // Retrieve several properties from the algorithm configuration.
        double horizontalSpaceBetweenTwoNodes = layoutGraph.getProperty(MindstormsOptions.SPACING_NODE_NODE);
        double verticalSpaceBetweenTwoLines = layoutGraph.getProperty(MindstormsOptions.SPACING_LINE);
        double nbNodesPerLine = layoutGraph.getProperty(MindstormsOptions.NODE_PLACEMENT_NB_NODES_PER_LINE);
        double verticalEdgeNodeSpacing = verticalSpaceBetweenTwoLines / 2;

        // Get the list of nodes to lay out
        List<ElkNode> nodes = new ArrayList<>(layoutGraph.getChildren());

        // The horizontal initial location of the first node of each line. 
        int initialXLocation = 20;
        if (nodes.size() > nbNodesPerLine) {
        	// Add a space on the left of all nodes for the edges loop
        	initialXLocation += HORIZONTAL_EGDE_LOOP_SPACING;
        }
        // The vertical initial location of the first node of each line. It will be incremented at each line.
        int initialYLocation = 20;
        
        // Create a sub monitor for node sort
        IElkProgressMonitor nodeSortingMonitor = progressMonitor.subTask(1);
        nodeSortingMonitor.begin("Node Sorting", 1);
        // Sort the list of nodes to lay out
        sortNodesAccordingToChoreography(nodes); 
        nodeSortingMonitor.done();
        progressMonitor.log("Node sorting done!");
        
        // Create a sub monitor for node placement
        IElkProgressMonitor nodePlacingMonitor = progressMonitor.subTask(1);
        nodePlacingMonitor.begin("Node Placing", nodes.size());
        
        // Place the nodes
        double currentLeftX = initialXLocation;
        double maxXRightLocation = initialXLocation;
        double currentTopY = initialYLocation;
        double currentBottomY = initialYLocation;
        
        // Make an output to the debug log
        nodePlacingMonitor.logGraph(layoutGraph, "No node placed yet");
        
        int currentIndexOfNodeInLine = 0;
        for (ElkNode node : nodes) {
        	if (currentIndexOfNodeInLine == nbNodesPerLine) {
        		maxXRightLocation = Math.max(maxXRightLocation, currentLeftX);
        		// Go to a new line
        		currentIndexOfNodeInLine = 0;
        		// x is reset to the initialXLocation
        		currentLeftX = initialXLocation;
        		// y is incremented according to the vertical space between two lines
        		currentTopY = currentBottomY + verticalSpaceBetweenTwoLines;
        		
        	}
            // Set the node's coordinates
        	nodePlacingMonitor.log(node.getIdentifier() + ": {" + currentLeftX + ", " + currentTopY + "}");
            node.setX(currentLeftX);
            node.setY(currentTopY);
            nodePlacingMonitor.logGraph(layoutGraph, node.getIdentifier() + " placed");
            
            // x is incremented according to the horizontal space between two nodes and the width of the current node
            currentLeftX += node.getWidth() + horizontalSpaceBetweenTwoNodes;
            currentBottomY = Math.max(currentBottomY, currentTopY + node.getHeight());
            currentIndexOfNodeInLine++;
        }
        
        if (!nodes.isEmpty()) {
        	maxXRightLocation = Math.max(maxXRightLocation, currentLeftX);
        	maxXRightLocation = maxXRightLocation - horizontalSpaceBetweenTwoNodes + HORIZONTAL_EGDE_LOOP_SPACING;
        }
        
        // Close the sub monitor
        nodePlacingMonitor.done();
        progressMonitor.log("Node Placing done!");
        
        // Create sub monitor for edge routing
        IElkProgressMonitor edgeRoutingMonitor = progressMonitor.subTask(1);
        edgeRoutingMonitor.begin("Edge Routing", layoutGraph.getContainedEdges().size());
        edgeRoutingMonitor.logGraph(layoutGraph, "No edge routed yet");
        
        // Route the edges
        if (!layoutGraph.getContainedEdges().isEmpty()) {
            for (ElkEdge edge : layoutGraph.getContainedEdges()) {
                ElkNode source = ElkGraphUtil.connectableShapeToNode(edge.getSources().get(0));
                ElkNode target = ElkGraphUtil.connectableShapeToNode(edge.getTargets().get(0));
                
                ElkEdgeSection section = ElkGraphUtil.firstEdgeSection(edge, true, true);
                
                section.setStartLocation(
                        source.getX() + source.getWidth(),
                        source.getY() + SPACE_FROM_TOP_OF_NODE);
                section.setEndLocation(
                        target.getX(),
                        target.getY() + SPACE_FROM_TOP_OF_NODE);

                if (source.getY() == target.getY()) {
                	edgeRoutingMonitor.log("Edge between 2 nodes of same line, from " + section.getStartX() + " to " + section.getEndX() + " on y axis " + section.getStartY());
                } else {
                	// Add bendpoints to the edge for the loop to going to a new line
                	ElkGraphUtil.createBendPoint(section, section.getStartX() + HORIZONTAL_EGDE_LOOP_SPACING, section.getStartY());
                	ElkGraphUtil.createBendPoint(section, section.getStartX() + HORIZONTAL_EGDE_LOOP_SPACING, section.getEndY() - SPACE_FROM_TOP_OF_NODE - verticalEdgeNodeSpacing);
                	ElkGraphUtil.createBendPoint(section, section.getEndX() - HORIZONTAL_EGDE_LOOP_SPACING, section.getEndY() - SPACE_FROM_TOP_OF_NODE - verticalEdgeNodeSpacing);
                	ElkGraphUtil.createBendPoint(section, section.getEndX() - HORIZONTAL_EGDE_LOOP_SPACING, section.getEndY());
                	edgeRoutingMonitor.log("Edge between 2 lines, with a loop, from {" + section.getStartX() + ", " + section.getStartY() + "} to {" + section.getEndX() + ", " + section.getEndY() + "}.");
                }
                edgeRoutingMonitor.logGraph(layoutGraph, source.getIdentifier() + " -> " + target.getIdentifier());
            }
        }
        
        // Close the sub monitor
        edgeRoutingMonitor.done();
        
        progressMonitor.log("Edge Routing done!");
        
        // Set the size of the final diagram
        layoutGraph.setWidth(maxXRightLocation);
        layoutGraph.setHeight(currentBottomY);
        
        layoutGraph.setProperty(CoreOptions.EDGE_ROUTING, EdgeRouting.ORTHOGONAL);
        
        // End the progress monitor
        progressMonitor.log("Algorithm executed");
        progressMonitor.logGraph(layoutGraph, "Final graph");
        progressMonitor.done();
    }

    /**
     * Sort the nodes according to flow defined in the choreography.
     * @param nodes List of nodes to be sorted
     */
    private void sortNodesAccordingToChoreography(List<ElkNode> nodes) {
    	Collections.sort(nodes, (node1, node2) -> {
    		int result = 0;
    		int incoming = 1;
    		int outgoing = -1;
    		if (node1.getIncomingEdges().stream().anyMatch(e -> ElkGraphUtil.getSourceNode(e) == node2)) {
    			result = incoming;
    		} else if (node1.getOutgoingEdges().stream().anyMatch(e -> ElkGraphUtil.getTargetNode(e) == node2)) {
    			result = outgoing;
    		} else {
    			// no direct link, search deeper...
    			List<ElkNode> incomingsNode = getAllIncomingNodes(node1, false);
    			if (incomingsNode.contains(node2)) {
    				result = incoming;
    			} else {
    				result = outgoing;
    			}
    		}
    		return result;
    	});
	}
    
    private List<ElkNode> getAllIncomingNodes(ElkNode node, boolean directIncoming) {
    	List<ElkNode> nodes = new ArrayList<>();
    	for (ElkEdge edge : node.getIncomingEdges()) {
    		ElkNode incomingNode = ElkGraphUtil.getSourceNode(edge);
    		if (directIncoming) {
    			nodes.add(incomingNode);
    		}
    		nodes.addAll(getAllIncomingNodes(incomingNode, true));
		}
		return nodes;
	}
}
